/*
 * fwflash.c - handling the betty flashes
 *
 * author: hackbard@hackdaworld.org
 *
 */

/*
 * include files
 */

#include <stdio.h>

#include "lpc2xxx.h"

/*
 * defines
 */

/* bank 0/2 and boootloader addr/size */
#define BANK0			0x80000000
#define BANK2			0x82000000
#define BANK_SIZE		0x00100000
#define BOOTLOADER		0x7fffe000
#define BL_SIZE			0x00002000

/* flash cmd addresses - flash[0-18] <--> arm[1-19]*/
#define B0F555	(*((volatile unsigned long *)(BANK0|0xaaa)))	// 0x555
#define B0F2AA	(*((volatile unsigned long *)(BANK0|0x554)))	// 0x2aa
#define B2F555	(*((volatile unsigned long *)(BANK2|0xaaa)))	// 0x555
#define B2F2AA	(*((volatile unsigned long *)(BANK2|0x554)))	// 0x2aa

/* commands */
#define CMD_READ		'R'
#define CMD_CHIP_ERASE		'E'
#define CMD_CHIP_ID		'I'

#define BUFSIZE			16

/*
 * type definitions
 */

typedef unsigned char u8;
typedef unsigned short u16;
typedef unsigned int u32;

/*
 * functions
 */

void mmap_init(u8 memtype) {

	MEMMAP=memtype;
}

void pll_init(void) {

	/* configuration */
	PLLCFG=0x02;		// multiplier = 2
	PLLCON=0x03;		// enable and set as clk source for the lpc
	/* feed sequence */
	PLLFEED=0xaa;
	PLLFEED=0x55;
	/* wait for lock */
	while(!(PLLSTAT&(1<<10)))
		continue;
}

void uart0_init(void) {

	PINSEL0=0x05;			// pin select -> tx, rx
	UART0_FCR=0x07;			// enable fifo
	UART0_LCR=0x83;			// set dlab + word length
	UART0_DLL=0x04;			// br: 38400 @ 10/4 mhz
	UART0_DLM=0x00;
	UART0_LCR=0x03;			// unset dlab
}

void uart0_send_string(char *txbuf) {

	int i;

	i=0;

	while(txbuf[i]) {
		UART0_THR=txbuf[i++];
		/* flush if tx buffer maximum reached */
		if(!(i%16))
			while(!(UART0_LSR&(1<<6)))
				continue;
	}
	
	/* flush if \n and \r do not fit in the tx buffer */
	if(i>14)
		while(!(UART0_LSR&(1<<6)))
			continue;

	UART0_THR='\n';
	UART0_THR='\r';

	/* flush uart0 anyways */
	while(!(UART0_LSR&(1<<6)))
		continue;
}

void uart0_send_buf16(u16 *buf,int len) {

	int i;

	i=0;

	for(i=0;i<len/2;i++) {
		if(!(i%8))
			while(!(UART0_LSR&(1<<6)))
				continue;
		UART0_THR=buf[i]&0xff;
		UART0_THR=(buf[i]>>8)&0xff;
	}
}

void uart0_send_buf32(u32 *buf,int len) {

	int i;

	i=0;

	for(i=0;i<len/4;i++) {
		if(!(i%4))
			while(!(UART0_LSR&(1<<6)))
				continue;
		UART0_THR=buf[i]&0xff;
		UART0_THR=(buf[i]>>8)&0xff;
		UART0_THR=(buf[i]>>16)&0xff;
		UART0_THR=(buf[i]>>24)&0xff;
	}
}

void uart0_send_byte(u8 send) {

	while(!(UART0_LSR&(1<<5)))
		continue;

	UART0_THR=send;
}

u8 uart0_get_byte(void) {

	u8 rx;

	while(!(UART0_LSR&(1<<0)))
		continue;

	rx=UART0_RBR;

	return rx;
}

/*
 * main function
 */

int main(void) {

	/* variables */

	u32 i,addrlen,datalen,addr;
	u8 buf[BUFSIZE];
	u16 data;
	u8 cmd;
	u8 txrx;

	/* memory mapping of interrupt vectors to static ram */

	//mmap_init(MMAP_RAM);
	
	/* pll initialization */
	pll_init();

	/* uart initialization */
	uart0_init();

	/* external memory init */
	BCFG0=0x1000FBEF;		// BCFG2 should be fine as is

	/* begin the main loop */
	while(1) {

	/* receive cmd */
	while(1) {

		cmd=uart0_get_byte();
		txrx=1;
		switch(cmd) {
			case CMD_CHIP_ERASE:
				addrlen=0;
				datalen=1;
				break;
			case CMD_READ:
				addrlen=4;
				datalen=4;
				break;
			case CMD_CHIP_ID:
				addrlen=0;
				datalen=1;
				break;
			default:
				txrx=0;
				break;
		}

		if(txrx)
			break;
	}

	/* receive (only if there is) more data from uart0 */
	addr=0;
	for(i=0;i<addrlen;i++) {
		txrx=uart0_get_byte();
		addr|=(txrx<<((addrlen-i-1)*8));
	}

	for(i=0;i<datalen;i++)
		buf[i]=uart0_get_byte();

	/* process the cmd */
	switch(cmd) {
		case CMD_READ:
			/* data length to read */
			datalen=buf[0]<<24|buf[1]<<16|buf[2]<<8|buf[3];
			/* check addr and datalen */
			if((addr>=BANK0)&(addr+datalen<=BANK0+BANK_SIZE))
				uart0_send_buf16((u16 *)addr,datalen);
			if((addr>=BANK2)&(addr+datalen<=BANK2+BANK_SIZE))
				uart0_send_buf16((u16 *)addr,datalen);
			if((addr>=BOOTLOADER)&(addr+datalen<=BOOTLOADER+BL_SIZE))
				uart0_send_buf32((u32 *)addr,datalen);
			break;
		case CMD_CHIP_ERASE:
			if(buf[0]=='0') {
				B0F555=0xaa;
				B0F2AA=0x55;
				B0F555=0x80;
				B0F555=0xaa;
				B0F2AA=0x55;
				B0F555=0x10;
			}
			else if(buf[0]=='2') {
				B2F555=0xaa;
				B2F2AA=0x55;
				B2F555=0x80;
				B2F555=0xaa;
				B2F2AA=0x55;
				B2F555=0x10;
			}
			uart0_send_byte(buf[0]);
			break;
		case CMD_CHIP_ID:
			if(buf[0]=='0') {
				B0F555=0xaa;
				B0F2AA=0x55;
				B0F555=0x90;
				data=*((u16 *)BANK0);
				data=*((u16 *)(BANK0|0x200));
			}
			else if(buf[0]=='2') {
				B2F555=0xaa;
				B2F2AA=0x55;
				B2F555=0x90;
				data=*((u16 *)BANK2);
				data=*((u16 *)(BANK2|0x200));
			}
			uart0_send_buf16(&data,2);
			break;
		default:
			break;
	}
		
	}

	return 0;
}

